// Copyright  2007-2014, Johan Hedborg, Linkping University
//                        Andreas Robinson, Linkping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////////////////////
//		M A T R I X  3 X 3
// Desc: A Matrix3x3 class
// 

#pragma once

#include "Vector3.hpp"

namespace cvl{

	template<typename T>
	class Matrix3x3{

	public:
		Matrix3x3(void);

		// look into the possibility to have a union 
		// between _a00 and a data pointer data[0] ( to T data[9] )
		// look into xtl (class packing code tip from Erik J)


		Matrix3x3(	T _a00, T _a01, T _a02,
					T _a10, T _a11, T _a12,
					T _a20, T _a21, T _a22 );

		Matrix3x3( const T* pData );

		// type casting 
		template<class U>
		operator Matrix3x3<U>() const
		{
			return Matrix3x3<U>((U)a00,(U)a01,(U)a02,(U)a10,(U)a11,(U)a12,(U)a20,(U)a21,(U)a22);
		}

		void set(const Vector3<T>& row0, const Vector3<T>& row1, const Vector3<T>& row2);

		void diag( T _a00, T _a11, T _a22 );

		T determinant();

		Matrix3x3<T> transpose() const;

		Matrix3x3<T> inverse() const;

		// inline const Matrix3x3<T>& operator=(const Matrix3x3<T>& m);

		Matrix3x3<T> &operator+=(const Matrix3x3<T> &Mat);
		Matrix3x3<T> &operator-=(const Matrix3x3<T> &Mat);
		Matrix3x3<T> &operator*=(T sclar);
		Matrix3x3<T> &operator/=(T sclar);
		
		
		inline Matrix3x3<T> operator-() const;
		inline Matrix3x3<T> operator-( const Matrix3x3<T>& v ) const;

		// FIXME make one for + also

		// Matrix3x3<T> operator*(const Matrix3x3<T> &A, const Matrix3x3<T> &B);
		//friend inline cvlVector3D<T> operator*(const Matrix3x3<T> &A, const cvlVector3D<T> &v);
		//friend inline Vector3d operator*(const Matrix3x3 &A, const Vector3d &v);

		//friend inline cvlVector3D<T> operator*(const cvlVector3D<T> &v,const Matrix3x3<T> &A);


		union {
			struct {
				T a00,a01,a02,a10,a11,a12,a20,a21,a22;
			};		
			T		m[3][3];
			T		n[9];			// rename this to v maybe ??
		};	



		T &operator()( const int& e, const int& i )
		{ return m[e][i]; }

		// FIXME : examine why cant i have ref here...
		T operator()( const int& e, const int& i ) const
		{ return m[e][i]; }

		T &operator[](const int& i )
		{ return n[i]; }

		const T &operator[](const int& i ) const
		{ return n[i]; }


		T* getPtr(void) { return &n[0]; } 
		const T* getConstPtr(void) const { return &n[0]; } 


		Vector3<T> row(int index)	const
		{
			return Vector3<T>( m[index][0], m[index][1], m[index][2]);
		}

		Vector3<T> col(int index)	const 
		{
			return Vector3<T>( m[0][index], m[1][index], m[2][index]);
		}

	};

	template<typename T>
	void Matrix3x3<T>::set(const Vector3<T>& row0, const Vector3<T>& row1, const Vector3<T>& row2)
	{
		a00=row0.x; a01=row1.x; a02=row2.x;
		a10=row0.y; a11=row1.y; a12=row2.y;
		a20=row0.z; a21=row1.z; a22=row2.z;	
	}
	

	template<typename T>
	void Matrix3x3<T>::diag( T _a00, T _a11, T _a22 )
	{
		a00=_a00; a01=0; a02=0;
		a10=0; a11=_a11; a12=0;
		a20=0; a21=0; a22=_a22;
	}

	template<typename T>
	Matrix3x3<T>::Matrix3x3(void)
	{}

	template<typename T>
	Matrix3x3<T>::Matrix3x3( T _a00, T _a01, T _a02,
		T _a10, T _a11, T _a12,
		T _a20, T _a21, T _a22 )
	{
		a00=_a00; a01=_a01; a02=_a02;
		a10=_a10; a11=_a11; a12=_a12;
		a20=_a20; a21=_a21; a22=_a22;
	}
	
	template<typename T>
	Matrix3x3<T>::Matrix3x3( const T* pData )
	{
		a00=pData[0]; a01=pData[1]; a02=pData[2];
		a10=pData[3]; a11=pData[4]; a12=pData[5];
		a20=pData[6]; a21=pData[7]; a22=pData[8];
	}



	template <typename T>
	inline T Matrix3x3<T>::determinant()
	{
		return 
			a00*a11*a22 -
			a00*a21*a12 +
			a10*a21*a02 -
			a10*a01*a22 +
			a20*a01*a12 -
			a20*a11*a02;
	}




	template<typename T>
	inline Matrix3x3<T> Matrix3x3<T>::transpose() const
	{
		return Matrix3x3<T>(a00,a10,a20,a01,a11,a21,a02,a12,a22);
	}


	template< typename T>
	std::ostream &operator << (std::ostream &os, Matrix3x3<T> &v) 
	{
		os << "[  ";

		for (int e = 0; e < 3 ; e++)
		{
			for (int i = 0; i < 3; i++){
				if (v(e,i) > 0)
					os << " ";
				os << v(e,i) << "  " ;
			}
			if (e != 2)
				os << "\n   ";
		}
		os << "]";
		return os;
	}



	template <typename T>
	inline Matrix3x3<T> Matrix3x3<T>::inverse() const 
	{
		//T d=determinant();

		//d=(d==0) ? 1 : d;
  //
		//// FIX ME! some problem dont use it until its fixed!
		//printf("Matrix3x3<T>::inverse(), dont work yet!\n");exit(-1);

		//return Matrix3x3((a11*a22-a12*a21)/d,
		//				-(a01*a22-a02*a21)/d,
		//				-(a01*a22-a02*a11)/d,
		//				-(a10*a22-a12*a20)/d,
		//				 (a00*a22-a02*a20)/d,
		//				-(a00*a12-a02*a10)/d,
		//				 (a10*a21-a11*a20)/d,
		//				-(a00*a21-a01*a20)/d,
		//				 (a00*a11-a01*a10)/d);
	

		// Maple generation : 
		// with(LinearAlgebra): Mat := Matrix(3,3,symbol = m):
		// with(CodeGeneration): C(MatrixInverse(Mat), optimize, resultname = "ret_m",precision=double);
	
		Matrix3x3 ret;

		T t4 = m[2][0] * m[0][1];
		T t6 = m[2][0] * m[0][2];
		T t8 = m[1][0] * m[0][1];
		T t10 = m[1][0] * m[0][2];
		T t12 = m[0][0] * m[1][1];
		T t14 = m[0][0] * m[1][2];

		T t17 = (t4 * m[1][2] - t6 * m[1][1] - t8 * m[2][2] + t10 * m[2][1] + t12 * m[2][2] - t14 * m[2][1]);
		// should check so 1/t17 != 0	FIXME!!!
		//if (t17 < )
		//{
		//}

		t17 = 0.1e1 / t17;

		ret.m[0][0] = (m[1][1] * m[2][2] - m[1][2] * m[2][1]) * t17;
		ret.m[0][1] = -(m[0][1] * m[2][2] - m[0][2] * m[2][1]) * t17;
		ret.m[0][2] = (m[0][1] * m[1][2] - m[0][2] * m[1][1]) * t17;
		ret.m[1][0] = -(-m[2][0] * m[1][2] + m[1][0] * m[2][2]) * t17;
		ret.m[1][1] = (-t6 + m[0][0] * m[2][2]) * t17;
		ret.m[1][2] = -(-t10 + t14) * t17;
		ret.m[2][0] = (-m[2][0] * m[1][1] + m[1][0] * m[2][1]) * t17;
		ret.m[2][1] = -(-t4 + m[0][0] * m[2][1]) * t17;
		ret.m[2][2] = (-t8 + t12) * t17;

		return ret;
	}

	template <typename T>
	inline Matrix3x3<T> &Matrix3x3<T>::operator+=(const Matrix3x3<T> &m){
		a00+=m.a00;  a01+=m.a01;  a02+=m.a02;
		a10+=m.a10;  a11+=m.a11;  a12+=m.a12;
		a20+=m.a20;  a21+=m.a21;  a22+=m.a22;
		return *this;
	}


	template <typename T>
	inline Matrix3x3<T> &Matrix3x3<T>::operator-=(const Matrix3x3<T> &m){
		a00-=m.a00;  a01-=m.a01;  a02-=m.a02;
		a10-=m.a10;  a11-=m.a11;  a12-=m.a12;
		a20-=m.a20;  a21-=m.a21;  a22-=m.a22;
		return *this;
	}


	template <typename T>
	inline Matrix3x3<T> &Matrix3x3<T>::operator*=(T sclar)
	{
		a00*=sclar;  a01*=sclar;  a02*=sclar;
		a10*=sclar;  a11*=sclar;  a12*=sclar;
		a20*=sclar;  a21*=sclar;  a22*=sclar;
		return *this;
	}


	template <typename T>
	inline Matrix3x3<T> &Matrix3x3<T>::operator/=(T sclar)
	{
		a00/=sclar;  a01/=sclar;  a02/=sclar;
		a10/=sclar;  a11/=sclar;  a12/=sclar;
		a20/=sclar;  a21/=sclar;  a22/=sclar;
		return *this;
	}


	//	-Matrix3x3
	template <typename T>
	inline Matrix3x3<T> Matrix3x3<T>::operator-() const 
	{
		return Matrix3x3<T>( -a00, -a01, -a02, -a10, -a11, -a12, -a20, -a21, -a22 );
	}


	// operator -
	template<typename T>
	inline Matrix3x3<T> Matrix3x3<T>::operator-( const Matrix3x3<T>& v ) const
	{
		return Matrix3x3<T>( a00-v.a00, a01-v.a01, a02-v.a02, a10-v.a10, a11-v.a11, a12-v.a12, a20-v.a20, a21-v.a21, a22 -v.a22 );
	}

	typedef Matrix3x3<float> Matrix3x3F;
	typedef Matrix3x3<double> Matrix3x3D;
}
