// Copyright  2007-2014, Johan Hedborg, Linkping University
//                        Andreas Robinson, Linkping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////////////////////
//		M A T R I X  4 X 4
// Desc: A Matrix3x4 class
// 

#pragma once

#include "Vector3.hpp"
#include "Vector4.hpp"
#include "Matrix3x3.hpp"

namespace cvl{

	template<typename T>
	class Matrix3x4{
	public:
		Matrix3x4<T>(void);

		Matrix3x4<T>(
			T _a00, T _a01, T _a02, T _a03,
			T _a10, T _a11, T _a12, T _a13,
			T _a20, T _a21, T _a22, T _a23);

		Matrix3x4( const T* pData );

		// type casting 
		template<class U>
		operator Matrix3x4<U>() const
		{
			return Matrix3x4<U>((U)a00,(U)a01,(U)a02,(U)a03 ,(U)a10,(U)a11,(U)a12,(U)a13 ,(U)a20,(U)a21,(U)a22,(U)a23);
		}


		void diag( T _a00, T _a11, T _a22 );

		T determinant();
		Matrix3x4<T> transpose();
		Matrix3x4<T> inverse();

		Matrix3x4<T> &operator+=(const Matrix3x4<T> &Mat);
		Matrix3x4<T> &operator-=(const Matrix3x4<T> &Mat);
		Matrix3x4<T> &operator*=(T sclar);
		Matrix3x4<T> &operator/=(T sclar);

		inline Matrix3x4<T> operator-();
		inline Matrix3x4<T> operator-() const;
		inline Matrix3x4<T> operator-( const Matrix3x4<T>& v ) const;

		union {
			struct {
				T a00,a01,a02,a03,a10,a11,a12,a13,a20,a21,a22,a23;
			};		
			T		m[3][4];
			T		n[12];
		};	



		T &operator()( const int& e, const int& i )
		{ return this->m[e][i]; }

		T operator()( const int& e, const int& i ) const
		{ return this->m[e][i]; }

		T &operator[](const int& i )
		{ return this->n[i]; }

		T operator[](const int& i ) const
		{ return this->n[i]; }

		Vector3<T> row(int index)
		{
			return Vector4<T>( m[index][0], m[index][1], m[index][2], m[index][3]);
		}

		Vector3<T> col(int index)
		{
			return Vector3<T>( m[0][index], m[1][index], m[2][index]);
		}
	};

	template<typename T>
	void Matrix3x4<T>::diag( T _a00, T _a11, T _a22)
	{
		a00=_a00; a01=0;    a02=0;    a03=0;
		a10=0;    a11=_a11; a12=0;    a13=0;
		a20=0;    a21=0;    a22=_a22; a23=0;
	}

	template<typename T>
	Matrix3x4<T>::Matrix3x4(void)
	{
	}

	template<typename T>
	Matrix3x4<T>::Matrix3x4(T _a00, T _a01, T _a02, T _a03,
							T _a10, T _a11, T _a12, T _a13,
							T _a20, T _a21, T _a22, T _a23  )
	{
			a00 = _a00; a01 = _a01; a02 = _a02 ; a03 = _a03;
			a10 = _a10, a11 = _a11; a12 = _a12 ; a13 = _a13,
			a20 = _a20, a21 = _a21; a22 = _a22 ; a23 = _a23;
	}


	template<typename T>
	Matrix3x4<T>::Matrix3x4( const T* pData )
	{
		a00=pData[0]; a01=pData[1]; a02=pData[2]; a03=pData[3]; 
		a10=pData[4]; a11=pData[5]; a12=pData[6]; a13=pData[7]; 
		a20=pData[8]; a21=pData[9]; a22=pData[10]; a23=pData[11]; 
	}


	template<typename T>
	inline Matrix3x4<T> Matrix3x4<T>::transpose()
	{
		return Matrix3x4<T>(a00,a10,a20,a01,a11,a21,a02,a12,a22,a03,a13,a23);
	}

	// 
	template<typename T>
	std::ostream& operator<<(std::ostream& out, Matrix3x4<T>& v) 
	{
		out << v.a00 << ", " << v.a01 << ", " << v.a02 << ", " << v.a03 << "\n" <<
			v.a10 << ", " << v.a11 << ", " << v.a12 << ", " << v.a13 << "\n" <<
			v.a20 << ", " << v.a21 << ", " << v.a22 << ", " << v.a23 << "\n";

		return out;
	}


	template <typename T>
	inline Matrix3x4<T> &Matrix3x4<T>::operator+=(const Matrix3x4<T> &m){
		a00+=m.a00;  a01+=m.a01;  a02+=m.a02;  a03+=m.a03;
		a10+=m.a10;  a11+=m.a11;  a12+=m.a12;  a13+=m.a13; 
		a20+=m.a20;  a21+=m.a21;  a22+=m.a22;  a23+=m.a23;
		return *this;
	}


	template <typename T>
	inline Matrix3x4<T> &Matrix3x4<T>::operator-=(const Matrix3x4<T> &m){
		a00-=m.a00;  a01-=m.a01;  a02-=m.a02;  a03-=m.a03;
		a10-=m.a10;  a11-=m.a11;  a12-=m.a12;  a13-=m.a13;
		a20-=m.a20;  a21-=m.a21;  a22-=m.a22;  a23-=m.a23;
		return *this;
	}

	//	-Matrix3x3
	template <typename T>
	inline Matrix3x4<T> Matrix3x4<T>::operator-()
	{
		return Matrix3x4<T>( -a00, -a01, -a02, -a03,  -a10, -a11, -a12, -a13,  -a20, -a21, -a22, -a23 );
	}

	template <typename T>
	inline Matrix3x4<T> Matrix3x4<T>::operator-() const 
	{
		return Matrix3x4<T>( -a00, -a01, -a02, -a03, -a10, -a11, -a12, -a13, -a20, -a21, -a22, -a23 );
	}

	template<typename T>
	inline Matrix3x4<T> Matrix3x4<T>::operator-( const Matrix3x4<T>& v ) const
	{
		return Matrix3x4<T>( a00-v.a00, a01-v.a01, a02-v.a02, a03-v.a03, a10-v.a10, a11-v.a11, a12-v.a12, a13-v.a13, a20-v.a20, a21-v.a21, a22 -v.a22, a23 -v.a23 );
	}


	typedef Matrix3x4<float> Matrix3x4F;
	typedef Matrix3x4<double> Matrix3x4D;
}
