// Copyright  2007-2014, Johan Hedborg, Linkping University
//                        Andreas Robinson, Linkping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////////////////////
//		M A T R I X  4 X 4
// Desc: A Matrix4x4 class
// 

#pragma once

#include "Vector4.hpp"

namespace cvl{

	template<typename T>
	class Matrix4x4{

	public:
		Matrix4x4<T>(void);

		// lock in to the possibility to have a union 
		// between _a00 and a data pointer data[0] ( to T data[9] )
		// look into xtl (class packing code tip from Erik J)


		Matrix4x4<T>(	T _a00, T _a01, T _a02, T _a03,
						T _a10, T _a11, T _a12, T _a13,
						T _a20, T _a21, T _a22, T _a23,
						T _a30, T _a31, T _a32, T _a33 );
			
		Matrix4x4( const T* pData );

		// type casting 
		template<class U>
		operator Matrix4x4<U>() const
		{
			return Matrix4x4<U>((U)a00,(U)a01,(U)a02,(U)a03,(U)a10,(U)a11,(U)a12,(U)a13,(U)a20,(U)a21,(U)a22,(U)a23,(U)a30,(U)a31,(U)a32,(U)a33);
		}
		
		void set(const Vector4<T>& row0, const Vector4<T>& row1, const Vector4<T>& row2, const Vector4<T>& row3);
				
		void diag( T _a00, T _a11, T _a22 , T _a33 );

		T determinant();

		
		Matrix4x4<T> transpose() const;

		Matrix4x4<T> inverse() const;

		// inline const Matrix4x4<T>& operator=(const Matrix4x4<T>& m);

		Matrix4x4<T> &operator+=(const Matrix4x4<T> &Mat);
		Matrix4x4<T> &operator-=(const Matrix4x4<T> &Mat);
		Matrix4x4<T> &operator*=(T sclar);
		Matrix4x4<T> &operator/=(T sclar);

		inline Matrix4x4<T> operator-() const;
		inline Matrix4x4<T> operator-( const Matrix4x4<T>& v ) const;
		// FIXME make one for + also

		// Matrix4x4<T> operator*(const Matrix4x4<T> &A, const Matrix4x4<T> &B);
		//friend inline CvlVector3D<T> operator*(const Matrix4x4<T> &A, const CvlVector3D<T> &v);
		//friend inline Vector3d operator*(const Matrix4x4 &A, const Vector3d &v);

		//friend inline CvlVector3D<T> operator*(const CvlVector3D<T> &v,const Matrix4x4<T> &A);

		union {
			struct {
				T a00,a01,a02,a03,a10,a11,a12,a13,a20,a21,a22,a23,a30,a31,a32,a33;
			};		
			T		m[4][4];
			T		n[16];
		};	


		T &operator()( const int& e, const int& i )
		{ return this->m[e][i]; }

		// FIXME : examine why cant i have ref here...
		T operator()( const int& e, const int& i ) const
		{ return m[e][i]; }

		T &operator[](const int& i )
		{ return this->n[i]; }

		const T &operator[](const int& i ) const 
		{ return this->n[i]; }



		T* getPtr(void) { return &n[0]; } 
		const T* getConstPtr(void) const { return &n[0]; } 


		Vector4<T> row(int index)
		{
			return Vector4<T>( m[index][0], m[index][1], m[index][2], m[index][3]);
		}

		Vector4<T> col(int index)
		{
			return Vector4<T>( m[0][index], m[1][index], m[2][index], m[3][index]);
		}

	};

	template<typename T>
	void Matrix4x4<T>::set(const Vector4<T>& row0, const Vector4<T>& row1, const Vector4<T>& row2, const Vector4<T>& row3)
	{
		a00=row0.x; a01=row1.x; a02=row2.x; a03=row3.x;
		a10=row0.y; a11=row1.y; a12=row2.y; a13=row3.y;
		a20=row0.z; a21=row1.z; a22=row2.z;	a23=row3.z;	
		a30=row0.w; a31=row1.w; a32=row2.w;	a33=row3.w;	
	}

	template<typename T>
	void Matrix4x4<T>::diag( T _a00, T _a11, T _a22, T _a33 )
	{
		a00=_a00; a01=0;    a02=0;    a03=0;
		a10=0;    a11=_a11; a12=0;    a13=0;
		a20=0;    a21=0;    a22=_a22; a23=0;
		a30=0;    a31=0;    a32=0;    a33=_a33;
	}

	template<typename T>
	Matrix4x4<T>::Matrix4x4(void)
	{
	}

	template<typename T>
	Matrix4x4<T>::Matrix4x4(T _a00, T _a01, T _a02, T _a03,
							T _a10, T _a11, T _a12, T _a13,
							T _a20, T _a21, T _a22, T _a23,	
							T _a30, T _a31, T _a32, T _a33  )
	{
		a00=_a00; a01=_a01; a02=_a02; a03=_a03;
		a10=_a10; a11=_a11; a12=_a12; a13=_a13;
		a20=_a20; a21=_a21; a22=_a22; a23=_a23;
		a30=_a30; a31=_a31; a32=_a32; a33=_a33;

	}


	template<typename T>
	Matrix4x4<T>::Matrix4x4( const T* pData )
	{
		memcpy(&a00, pData, sizeof(T)*16);
	}


	// generated by maple by 
	// M := Matrix(4, 4, [ a00, a01, a02, a03, a10, a11, a12, a13, a20, a21, a22, a23, a30, a31, a32, a33 ])
	// with(CodeGeneration) : C(Determinant(M));
	template <typename T>
	inline T Matrix4x4<T>::determinant()
	{
		return a00 * a11 * a22 * a33 - a00 * a11 * a23 * a32 + a00 * a21 * a32 * a13 - a00 * a21 * a12 * a33 + a00 * a31 * a12 * a23 - a00 * a31 * a22 * a13 - 
			a10 * a01 * a22 * a33 + a10 * a01 * a23 * a32 - a10 * a21 * a32 * a03 + a10 * a21 * a02 * a33 - a10 * a31 * a02 * a23 + a10 * a31 * a22 * a03 + a20 * a01 * a12 * a33 - 
			a20 * a01 * a32 * a13 + a20 * a11 * a32 * a03 - a20 * a11 * a02 * a33 + a20 * a31 * a02 * a13 - a20 * a31 * a12 * a03 - a30 * a01 * a12 * a23 + a30 * a01 * a22 * a13 - 
			a30 * a11 * a22 * a03 + a30 * a11 * a02 * a23 - a30 * a21 * a02 * a13 + a30 * a21 * a12 * a03;

	}




	template<typename T>
	inline Matrix4x4<T> Matrix4x4<T>::transpose() const
	{
		return Matrix4x4<T>(a00,a10,a20,a30,a01,a11,a21,a31,a02,a12,a22,a32,a03,a13,a23,a33);
	}


	// obsolete, use operator << 
	//template<typename T>
	//std::string Matrix4x4<T>::toString()
	//{
	//	char str[256];
	//	sprintf(str, "%g,%g,%g,%g\n%g,%g,%g,%g\n%g,%g,%g,%g\n%g,%g,%g,%g",
	//		a00, a01, a02, a03,
	//		a10, a11, a12, a13,
	//		a20, a21, a22, a23,
	//		a30, a31, a32, a33 );
	//	return std::string(str);
	//}



	//Matrix Inversion using Cramer's Method
	//Computes Adjoint matrix divided by determinant
	// this can be optimized further !
	// by eliminate the copy to mat thingi
	template <typename T>
	inline Matrix4x4<T> Matrix4x4<T>::inverse() const
	{
		//T* mat = new T[16];
		//T* dst = new T[16];
		//T* tmp = new T[12];
		//T* src = new T[16];
		T mat[16];
		T dst[16];
		T tmp[12];
		T src[16];

		for(int i=0; i<16; i++)
			mat[i] = (&a00)[i];



		// arrayoftransposesourcematrix
		T det;

		// determinant
		// transposematrix
		for(int i=0;i<4;i++) //>
		{
			src[i]=mat[i*4];
			src[i+4]=mat[i*4+1];
			src[i+8]=mat[i*4+2];
			src[i+12]=mat[i*4+3];
		}

		// calculatepairsforfirst8elements(cofactors)
		tmp[0]=src[10]*src[15];
		tmp[1]=src[11]*src[14];
		tmp[2]=src[9]*src[15];
		tmp[3]=src[11]*src[13];
		tmp[4]=src[9]*src[14];
		tmp[5]=src[10]*src[13];
		tmp[6]=src[8]*src[15];
		tmp[7]=src[11]*src[12];
		tmp[8]=src[8]*src[14];
		tmp[9]=src[10]*src[12];
		tmp[10]=src[8]*src[13];
		tmp[11]=src[9]*src[12];

		// calculatefirst8elements(cofactors)
		dst[0]=tmp[0]*src[5]+tmp[3]*src[6]+tmp[4]*src[7];
		dst[0]-=tmp[1]*src[5]+tmp[2]*src[6]+tmp[5]*src[7];
		dst[1]=tmp[1]*src[4]+tmp[6]*src[6]+tmp[9]*src[7];
		dst[1]-=tmp[0]*src[4]+tmp[7]*src[6]+tmp[8]*src[7];
		dst[2]=tmp[2]*src[4]+tmp[7]*src[5]+tmp[10]*src[7];
		dst[2]-=tmp[3]*src[4]+tmp[6]*src[5]+tmp[11]*src[7];
		dst[3]=tmp[5]*src[4]+tmp[8]*src[5]+tmp[11]*src[6];
		dst[3]-=tmp[4]*src[4]+tmp[9]*src[5]+tmp[10]*src[6];
		dst[4]=tmp[1]*src[1]+tmp[2]*src[2]+tmp[5]*src[3];
		dst[4]-=tmp[0]*src[1]+tmp[3]*src[2]+tmp[4]*src[3];
		dst[5]=tmp[0]*src[0]+tmp[7]*src[2]+tmp[8]*src[3];
		dst[5]-=tmp[1]*src[0]+tmp[6]*src[2]+tmp[9]*src[3];
		dst[6]=tmp[3]*src[0]+tmp[6]*src[1]+tmp[11]*src[3];
		dst[6]-=tmp[2]*src[0]+tmp[7]*src[1]+tmp[10]*src[3];
		dst[7]=tmp[4]*src[0]+tmp[9]*src[1]+tmp[10]*src[2];
		dst[7]-=tmp[5]*src[0]+tmp[8]*src[1]+tmp[11]*src[2];

		//calculatepairsforsecond8elements(cofactors)
		tmp[0]=src[2]*src[7];
		tmp[1]=src[3]*src[6];
		tmp[2]=src[1]*src[7];
		tmp[3]=src[3]*src[5];
		tmp[4]=src[1]*src[6];
		tmp[5]=src[2]*src[5];
		tmp[6]=src[0]*src[7];
		tmp[7]=src[3]*src[4];
		tmp[8]=src[0]*src[6];
		tmp[9]=src[2]*src[4];
		tmp[10]=src[0]*src[5];
		tmp[11]=src[1]*src[4];

		// calculatesecond8elements(cofactors)
		dst[8]=tmp[0]*src[13]+tmp[3]*src[14]+tmp[4]*src[15];
		dst[8]-=tmp[1]*src[13]+tmp[2]*src[14]+tmp[5]*src[15];
		dst[9]=tmp[1]*src[12]+tmp[6]*src[14]+tmp[9]*src[15];
		dst[9]-=tmp[0]*src[12]+tmp[7]*src[14]+tmp[8]*src[15];
		dst[10]=tmp[2]*src[12]+tmp[7]*src[13]+tmp[10]*src[15];
		dst[10]-=tmp[3]*src[12]+tmp[6]*src[13]+tmp[11]*src[15];
		dst[11]=tmp[5]*src[12]+tmp[8]*src[13]+tmp[11]*src[14];
		dst[11]-=tmp[4]*src[12]+tmp[9]*src[13]+tmp[10]*src[14];
		dst[12]=tmp[2]*src[10]+tmp[5]*src[11]+tmp[1]*src[9];
		dst[12]-=tmp[4]*src[11]+tmp[0]*src[9]+tmp[3]*src[10];
		dst[13]=tmp[8]*src[11]+tmp[0]*src[8]+tmp[7]*src[10];
		dst[13]-=tmp[6]*src[10]+tmp[9]*src[11]+tmp[1]*src[8];
		dst[14]=tmp[6]*src[9]+tmp[11]*src[11]+tmp[3]*src[8];
		dst[14]-=tmp[10]*src[11]+tmp[2]*src[8]+tmp[7]*src[9];
		dst[15]=tmp[10]*src[10]+tmp[4]*src[8]+tmp[9]*src[9];
		dst[15]-=tmp[8]*src[9]+tmp[11]*src[10]+tmp[5]*src[8];

		// calculate determinant
		det=src[0]*dst[0]+src[1]*dst[1]+src[2]*dst[2]+src[3]*dst[3];

		// calculate matrix inverse
		det=1/det;

		for(int j=0;j<16;j++) 
			dst[j]*=det;

		// Copy everything into the output
		Matrix4x4 ret;
		//k = 0;
		//for(int i=0; i<4; i++)
		//	for(int j=0; j<4; j++)
		//		ret.matrix[i][j] = dst[k++];

		for(int i=0; i<16; i++)
			(&ret.a00)[i] = dst[i];


		//delete[] mat;
		//delete[] dst;
		//delete[] tmp;
		//delete[] src;


		return ret;
	}



	template <typename T>
	inline Matrix4x4<T> &Matrix4x4<T>::operator+=(const Matrix4x4<T> &m)
	{
		a00+=m.a00;  a01+=m.a01;  a02+=m.a02;  a03+=m.a03;
		a10+=m.a10;  a11+=m.a11;  a12+=m.a12;  a13+=m.a13; 
		a20+=m.a20;  a21+=m.a21;  a22+=m.a22;  a23+=m.a23;
		a30+=m.a30;  a31+=m.a31;  a32+=m.a32;  a33+=m.a33;
		return *this;
	}


	template <typename T>
	inline Matrix4x4<T> &Matrix4x4<T>::operator-=(const Matrix4x4<T> &m)
	{
		a00-=m.a00;  a01-=m.a01;  a02-=m.a02;  a03-=m.a03;
		a10-=m.a10;  a11-=m.a11;  a12-=m.a12;  a13-=m.a13;
		a20-=m.a20;  a21-=m.a21;  a22-=m.a22;  a23-=m.a23;
		a30-=m.a30;  a31-=m.a31;  a32-=m.a32;  a33-=m.a33;
		return *this;
	}




	//	-Matrix4x4
	template <typename T>
	inline Matrix4x4<T> Matrix4x4<T>::operator-() const 
	{
		return Matrix4x4<T>( -a00, -a01, -a02, -a03, -a10, -a11, -a12, -a13, -a20, -a21, -a22, -a23 , -a30, -a31, -a32, -a33 );
	}

	// operator -
	template<typename T>
	inline Matrix4x4<T> Matrix4x4<T>::operator-( const Matrix4x4<T>& v ) const
	{
		return Matrix4x4<T>( a00-v.a00, a01-v.a01, a02-v.a02, a03-v.a03, a10-v.a10, a11-v.a11, a12-v.a12, a13-v.a13, a20-v.a20, a21-v.a21, a22 -v.a22 , a23 -v.a23 , a30-v.a30, a31-v.a31, a32 -v.a32 , a33 -v.a33 );
	}


	typedef Matrix4x4<float> Matrix4x4F;
	typedef Matrix4x4<double> Matrix4x4D;




}
