// Copyright © 2007-2014, Johan Hedborg, Linköping University
//                        Andreas Robinson, Linköping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#pragma once

#include "Vector2.hpp"
#include "Vector3.hpp"
#include "Matrix3x4.hpp"
#include "MathUtilities.h"
#include "PowellDogLeg.hpp"

class Triangulator
{
public:

	DogLegTriangulator<double,3,6> solver;

	cvl::Vector3D calc( const std::vector<cvl::Matrix3x4D>& Ps, const std::vector<cvl::Vector2D>& imgPoint );

private:

	std::vector<cvl::Matrix3x4D> m_Ps;
	std::vector<cvl::Vector2D> m_imgPoint;
	double xInit[3];

	int LineLineIntersect( cvl::Vector3D& p1, cvl::Vector3D& p2, cvl::Vector3D& p3, cvl::Vector3D& p4, cvl::Vector3D& pa, cvl::Vector3D& pb, double& mua, double& mub );
	int TriangulateMidpoint( const double* C0, const double* C1, const double* img0Point2D, const cvl::Vector2D& img1Point2D , double* resultPoint3D );
	cvl::Vector3D TriangulateMidpoint( const cvl::Matrix3x4D& C0, const cvl::Matrix3x4D& C1, const cvl::Vector2D& x0, const cvl::Vector2D& x1 );
	void calc( cvl::Vector3D& xOpt );
};
