// Copyright © 2007-2014, Johan Hedborg, Linköping University
//                        Andreas Robinson, Linköping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////////////////////
//		printMatrixT 
// Desc : a file for the template base functions of print matrix, this is to avoid linkage errors

#pragma once

#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <iostream>


// print a MxN matrix, where an element a_ij lyes in row i and colom j
//  iStart,iStop > 0 and iStart,iStop < M
//  jStart,jStop > 0 and jStart,jStop < N
template<class T>
void printMatrixT(const T* pMatrix, int N, int iStart, int iStop, int jStart, int jStop )
{
	// std::cout can be configered 
	//std::cout.precision(4);
	//std::cout.setf(std::ios::fixed, std::ios::floatfield);


	assert(pMatrix);

	// get maxAbsValue = max(max(abs(pMatrix)))
	double maxValueToPrint = fabs(pMatrix[0]);
	for (int i = iStart; i <= iStop; i++)
		for (int j = jStart; j <= jStop; j++)
			if ( maxValueToPrint < fabs(pMatrix[ N * i + j ]) )
				maxValueToPrint = fabs(pMatrix[ N * i + j ]);

	// get the number of digits 
	int nrDigits = (int(log10(fabs(maxValueToPrint)))+1);

	//printf("\n");
	for (int i = iStart; i <= iStop; i++)
	{
		for (int j = jStart; j <= jStop; j++)
		{
			T pixelRenderd = pMatrix[ N * i + j ];

			// needed here due to strange printf behavior	
			if ( pixelRenderd == -0 )
				pixelRenderd = 0;

			// space between the columns (correct for the '-')
			if ( pixelRenderd >= 0 )
				printf("  ");
			else 
				printf(" ");


			// fill with ' '
			int l_nrDigits = 1;
			if (pixelRenderd != 0 && fabs(pixelRenderd) >= 1.0)
				l_nrDigits = (int( log10( fabs(pixelRenderd) ) )+1);
			for (int spaceNo = l_nrDigits; spaceNo < nrDigits ; spaceNo++)
				printf(" ");

			std::cout<<pixelRenderd;

		}
		if(i != iStop)		// not for the last row
			printf("\n");
	}
}

template<class T>
void printMatrix(const char* preStr, const T* pMatrix, const int M, const int N, const char* postStr)
{
	// std::cout can be configered 
	std::cout.precision(4);
	std::cout.setf(std::ios::fixed, std::ios::floatfield);
	
	printf("%s",preStr);
	printMatrix<T>( pMatrix, M, N );
	printf("%s",postStr);
}

template<class T>
void printMatrix(const T* pMatrix, const int M, const int N)
{
	std::cout.precision(4);
	std::cout.setf(std::ios::fixed, std::ios::floatfield);

	int maxNoRow = 8;
	if(N <= maxNoRow)
		printMatrixT<T>(pMatrix, N, 0,M-1, 0,N-1);
	else
	{
		printMatrixT<T>(pMatrix, N, 0,M-1, 0,maxNoRow-1);
		printf("\n\n");
		printMatrixT<T>(pMatrix, N, 0,M-1, maxNoRow,N-1);
	}
}






